/*
 * Copyright (c) 2021 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <aul.h>
#include <aul_job_scheduler.h>

#include <algorithm>

#include "job-manager.hh"
#include "log-private.hh"

namespace tizen_cpp {

void JobManager::AddJobHandler(JobHandler* handler) {
  handlers_.push_back(handler);
}

void JobManager::RemoveJobHandler(JobHandler* handler) {
  auto found = std::find(handlers_.begin(), handlers_.end(), handler);
  if (found == handlers_.end())
    return;

  handlers_.erase(found);
}

bool JobManager::ExistJobHandler(const std::string& job_id) {
  for (auto* h : handlers_) {
    if (h->GetJobId() == job_id)
      return true;
  }

  return false;
}

bool JobManager::ExistJobHandler(JobHandler* handler) {
  for (auto* h :  handlers_) {
    if (h == handler)
      return true;
  }

  return false;
}

void JobManager::Do(std::shared_ptr<Job> job) {
  job->SetEventListener(this);
  if (!ExistJobHandler(job->GetId())) {
    // LCOV_EXCL_START
    job->SetTimer();
    AddPendingJob(job);
    // LCOV_EXCL_STOP
  } else {
    job->SetIdler();
    AddRunningJob(job);
  }
}

int JobManager::Finish(const std::string& job_id) {
  RemoveRunningJob(job_id);
  int ret = aul_job_scheduler_update_job_status(job_id.c_str(),
      JOB_STATUS_FINISHED);
  if (ret != AUL_R_OK) {
    // LCOV_EXCL_START
    _E("aul_job_scheduler_update_job_status() is failed. error(%d)", ret);
    return ret;
    // LCOV_EXCL_STOP
  }

  return 0;
}

void JobManager::FinishAllJobs() {
  running_jobs_.clear();
  pending_jobs_.clear();
}

void JobManager::FlushPendingJob(const std::string& job_id) {
  auto iter = pending_jobs_.begin();
  while (iter != pending_jobs_.end()) {
    // LCOV_EXCL_START
    if ((*iter)->GetId() == job_id) {
      auto job = *iter;
      iter = pending_jobs_.erase(iter);
      job->UnsetTimer();
      AddPendingJob(job);
      job->SetIdler();
    // LCOV_EXCL_STOP
    } else {
      iter++;
    }
  }
}

void JobManager::AddRunningJob(std::shared_ptr<Job> job) {
  running_jobs_.push_back(job);
}

void JobManager::RemoveRunningJob(const std::string& job_id) {
  auto iter = running_jobs_.begin();
  while (iter != running_jobs_.end()) {
    if ((*iter)->GetId() == job_id)
      iter = running_jobs_.erase(iter);
    else
      iter++;
  }
}

// LCOV_EXCL_START
void JobManager::AddPendingJob(std::shared_ptr<Job> job) {
  pending_jobs_.push_back(job);
}
// LCOV_EXCL_STOP

// LCOV_EXCL_START
void JobManager::RemovePendingJob(const std::string& job_id) {
  auto iter = pending_jobs_.begin();
  while (iter != pending_jobs_.end()) {
    if ((*iter)->GetId() == job_id)
      iter = pending_jobs_.erase(iter);
    else
      iter++;
  }
}
// LCOV_EXCL_STOP

// LCOV_EXCL_START
void JobManager::OnTimedOut(const Job* job) {
  _E("[__TIMEDOUT__] Job(%s), Status(%d)",
      job->GetId().c_str(), job->GetStatus());
  RemovePendingJob(job->GetId());
}
// LCOV_EXCL_STOP

void JobManager::OnRun(const Job* job) {
  _D("[__JOB__] START. Job(%s), Status(%d)",
      job->GetId().c_str(), job->GetStatus());
  if (job->GetStatus() == SERVICE_APP_JOB_STATUS_START) {
    aul_job_scheduler_update_job_status(job->GetId().c_str(), JOB_STATUS_START);
    for (auto* h : handlers_) {
      if (h->GetJobId() == job->GetId())
        h->Do(job);
    }
  } else {
    // LCOV_EXCL_START
    for (auto* h : handlers_) {
      if (h->GetJobId() == job->GetId())
        h->Do(job);
    }
    aul_job_scheduler_update_job_status(job->GetId().c_str(),
        JOB_STATUS_STOPPED);
    // LCOV_EXCL_STOP
  }
  _D("[__JOB__] END. Job(%s), Status(%d)",
      job->GetId().c_str(), job->GetStatus());
  RemoveRunningJob(job->GetId());
}

}  // namespace tizen_cpp
