/*
 * Copyright (c) 2021 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "job.hh"
#include "log-private.hh"

namespace tizen_cpp {

Job::Job(int status, std::string id, tizen_base::Bundle data)
    : status_(status), id_(std::move(id)), data_(std::move(data)) {
}

Job::~Job() {
  UnsetIdler();
  UnsetTimer();
}

int Job::GetStatus() const {
  return status_;
}

const std::string& Job::GetId() const {
  return id_;
}

bundle* Job::GetData() const {
  return data_.GetHandle();
}

void Job::SetEventListener(IEvent* listener) {
  listener_ = listener;
}

// LCOV_EXCL_START
void Job::SetTimer() {
  if (timer_ == 0) {
    timer_ = g_timeout_add(5000, TimedOutCb, this);
    if (timer_ == 0)
      _E("g_timeout_add() is failed");
  }
}
// LCOV_EXCL_STOP

void Job::UnsetTimer() {
  if (timer_ != 0) {
    // LCOV_EXCL_START
    g_source_remove(timer_);
    timer_ = 0;
    // LCOV_EXCL_STOP
  }
}

void Job::SetIdler() {
  if (idler_ == 0) {
    idler_ = g_idle_add(IdleCb, this);
    if (idler_ == 0)
      _E("g_idle_add() is failed");  // LCOV_EXCL_LINE
  }
}

void Job::UnsetIdler() {
  if (idler_ != 0) {
    // LCOV_EXCL_START
    g_source_remove(idler_);
    idler_ = 0;
    // LCOV_EXCL_STOP
  }
}

// LCOV_EXCL_START
gboolean Job::TimedOutCb(gpointer data) {
  auto* handle = static_cast<Job*>(data);
  handle->timer_ = 0;

  auto* listener = handle->listener_;
  if (listener != nullptr)
    listener->OnTimedOut(handle);

  return G_SOURCE_REMOVE;
}
// LCOV_EXCL_STOP

gboolean Job::IdleCb(gpointer data) {
  auto* handle = static_cast<Job*>(data);
  handle->idler_ = 0;
  auto* listener = handle->listener_;
  if (listener != nullptr)
    listener->OnRun(handle);

  return G_SOURCE_REMOVE;
}

}  // namespace tizen_cpp
