/*
 * Copyright (c) 2021 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Ecore.h>
#include <app_control_internal.h>
#include <aul.h>
#include <aul_job_scheduler.h>
#include <dlog.h>
#include <glib.h>
#include <service_app.h>
#include <stdlib.h>
#include <unistd.h>
#include <vconf-internal-keys.h>

#include <app_core_base.hh>
#include <app_event_internal.hh>

#include <list>
#include <memory>
#include <mutex>
#include <type_traits>

#include "service_app_extension.h"
#include "service_app_internal.h"

#include "job-manager.hh"
#include "log-private.hh"
#include <stdexcept>

using namespace tizen_cpp;

namespace {

const char* __error_to_string(app_error_e error) {
  switch (error) {
    case APP_ERROR_NONE:
      return "NONE";
    case APP_ERROR_INVALID_PARAMETER:
      return "INVALID_PARAMETER";
    case APP_ERROR_OUT_OF_MEMORY:  // LCOV_EXCL_LINE
      return "OUT_OF_MEMORY";  // LCOV_EXCL_LINE
    case APP_ERROR_INVALID_CONTEXT:
      return "INVALID_CONTEXT";
    case APP_ERROR_NO_SUCH_FILE:  // LCOV_EXCL_LINE
      return "NO_SUCH_FILE";  // LCOV_EXCL_LINE
    case APP_ERROR_ALREADY_RUNNING:
      return "ALREADY_RUNNING";
    default:  // LCOV_EXCL_LINE
      return "UNKNOWN";  // LCOV_EXCL_LINE
  }
}

int __on_error(app_error_e error, const char* function,
    const char* description) {
  if (description) {
    _E("[%s] %s(0x%08x) : %s",
        function, __error_to_string(error), error, description);
  } else {
    _E("[%s] %s(0x%08x)", function, __error_to_string(error), error);
  }

  return error;
}

constexpr int APP_EVENT_MAX = static_cast<int>(IAppCore::IEvent::Type::END);
constexpr IAppCore::IEvent::Type __app_event_converter[APP_EVENT_MAX] = {
    [APP_EVENT_LOW_MEMORY] = IAppCore::IEvent::Type::LOW_MEMORY,
    [APP_EVENT_LOW_BATTERY] = IAppCore::IEvent::Type::LOW_BATTERY,
    [APP_EVENT_LANGUAGE_CHANGED] = IAppCore::IEvent::Type::LANG_CHANGE,
    [APP_EVENT_DEVICE_ORIENTATION_CHANGED] =
        IAppCore::IEvent::Type::DEVICE_ORIENTATION_CHANGED,
    [APP_EVENT_REGION_FORMAT_CHANGED] = IAppCore::IEvent::Type::REGION_CHANGE,
    [APP_EVENT_SUSPENDED_STATE_CHANGED] =
        IAppCore::IEvent::Type::SUSPENDED_STATE_CHANGE,
    [APP_EVENT_UPDATE_REQUESTED] = IAppCore::IEvent::Type::UPDATE_REQUESTED,
    [APP_EVENT_TIME_ZONE_CHANGED] =
        IAppCore::IEvent::Type::TIME_ZONE_CHANGED,
};

class AppContext : public AppCoreBase {
 public:
  enum AppState {
    APP_STATE_NOT_RUNNING,
    APP_STATE_CREATING,
    APP_STATE_RUNNING,
  };

  AppContext(service_app_lifecycle_callback_s callback,
      service_app_loop_method_s method, void* data)
          : callback_(callback), method_(method), data_(data) {
    state_ = APP_STATE_CREATING;
  }

  void Run(int argc, char** argv) override {
    SetAppState(APP_STATE_RUNNING);
    AppCoreBase::Run(argc, argv);
    SetAppState(APP_STATE_NOT_RUNNING);
  }

  int OnReceive(aul_type type, tizen_base::Bundle b) override {
    AppCoreBase::OnReceive(type, std::move(b));

    if (type == AUL_TERMINATE_BGAPP) {
      // LCOV_EXCL_START
      Exit();
      return 0;
      // LCOV_EXCL_STOP
    }

    return 0;
  }

  int OnCreate() override {
    _W("service_app_create()");
    AppCoreBase::OnCreate();

    if (callback_.create == nullptr || callback_.create(data_) == false) {
      return __on_error(APP_ERROR_INVALID_CONTEXT, __FUNCTION__,
          "service_app_create_cb() returns false");
    }

    return APP_ERROR_NONE;
  }

  int OnControl(tizen_base::Bundle b) override {
    _W("service_app_control()");
    AppCoreBase::OnControl(b);

    std::string job_id = b.GetString(AUL_K_JOB_ID);
    if (!job_id.empty()) {
      _D("[__JOB__] Job(%s)", job_id.c_str());  // LCOV_EXCL_LINE
      return 0;
    }

    app_control_h app_control = nullptr;
    if (app_control_create_event(b.GetHandle(), &app_control) != 0)
      return -1;

    if (callback_.app_control)
      callback_.app_control(app_control, data_);

    app_control_destroy(app_control);

    return 0;
  }

  int OnTerminate() override {
    _W("service_app_terminate()");
    if (callback_.terminate)
      callback_.terminate(data_);

    AppCoreBase::OnTerminate();

    return APP_ERROR_NONE;
  }

  void OnLoopInit(int argc, char** argv) override {
    if (method_.init) {
      method_.init(argc, argv, data_);  // LCOV_EXCL_LINE
    } else {
      _W("ecore_init");
      ecore_init();
    }
  }

  void OnLoopFinish() override {
    if (method_.fini) {
      method_.fini();  // LCOV_EXCL_LINE
    } else {
      _W("ecore_shutdown");
      ecore_shutdown();
    }
  }

  void OnLoopRun() override {
    if (method_.run)
      method_.run(data_);  // LCOV_EXCL_LINE
    else
      ecore_main_loop_begin();
  }

  void OnLoopExit() override {
    SetAppState(APP_STATE_NOT_RUNNING);
    if (method_.exit) {
      method_.exit(data_);  // LCOV_EXCL_LINE
    } else {
      ecore_main_loop_thread_safe_call_sync([](void* data) -> void* {
        ecore_main_loop_quit();
        return nullptr;
      }, nullptr);
    }
  }

  AppState GetAppState() const {
    std::lock_guard<std::recursive_mutex> lock(mutex_);
    return state_;
  }

  void SetAppState(AppState state) {
    std::lock_guard<std::recursive_mutex> lock(mutex_);
    state_ = state;
  }

 private:
  service_app_lifecycle_callback_s callback_;
  service_app_loop_method_s method_;
  void* data_;
  AppState state_ = APP_STATE_NOT_RUNNING;
  mutable std::recursive_mutex mutex_;
};

std::unique_ptr<AppContext> __context;
std::list<std::shared_ptr<AppEvent>> __pending_app_events;
JobManager __job_manager;

}  // namespace

extern "C" EXPORT_API int service_app_main_ext(int argc, char** argv,
    service_app_lifecycle_callback_s* callback,
    service_app_loop_method_s* method, void* user_data) {
  if (argc < 1 || argv == nullptr || callback == nullptr || method == nullptr)
    return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, nullptr);

  if (callback->create == nullptr) {
    return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__,
        "service_app_create_cb() callback must be registered");
  }

  if (__context.get() != nullptr &&
      __context->GetAppState() != AppContext::APP_STATE_NOT_RUNNING)
    return __on_error(APP_ERROR_ALREADY_RUNNING, __FUNCTION__, nullptr);

  try {
    __context = std::make_unique<AppContext>(*callback, *method, user_data);
    for (auto& i : __pending_app_events) {
      __context->AddEvent(i);
    }
    __context->Run(argc, argv);
  } catch(std::runtime_error& e) {
    // LCOV_EXCL_START
    __context->SetAppState(AppContext::APP_STATE_NOT_RUNNING);
    return __on_error(APP_ERROR_INVALID_CONTEXT, __FUNCTION__, nullptr);
    // LCOV_EXCL_STOP
  }

  __job_manager.FinishAllJobs();
  return APP_ERROR_NONE;
}

extern "C" EXPORT_API int service_app_main(int argc, char** argv,
    service_app_lifecycle_callback_s* callback, void* user_data) {
  service_app_loop_method_s method = { nullptr, };
  _W("service_app_main()");
  return service_app_main_ext(argc, argv, callback, &method, user_data);
}

extern "C" EXPORT_API void service_app_exit(void) {
  _W("service_app_exit()");
  if (__context.get() &&
      __context->GetAppState() == AppContext::APP_STATE_RUNNING)
    __context->Exit();
}

extern "C" EXPORT_API int service_app_add_event_handler(
    app_event_handler_h* event_handler, app_event_type_e event_type,
    app_event_cb callback, void *user_data) {
  if (event_handler == nullptr || callback == nullptr) {
    return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__,
        "null parameter");
  }

  if (event_type < APP_EVENT_LOW_MEMORY ||
      event_type > APP_EVENT_TIME_ZONE_CHANGED ||
      event_type == APP_EVENT_DEVICE_ORIENTATION_CHANGED) {
    return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__,
        "invalid event type");
  }

  auto* ae = new (std::nothrow) AppEvent(__app_event_converter[event_type],
      callback, user_data);
  if (!ae) {
    // LCOV_EXCL_START
    return __on_error(APP_ERROR_OUT_OF_MEMORY, __FUNCTION__,
        "failed to create handler");
    // LCOV_EXCL_STOP
  }

  auto* h = new (std::nothrow) std::shared_ptr<AppEvent>(ae);
  if (!h) {
    // LCOV_EXCL_START
    delete ae;
    return __on_error(APP_ERROR_OUT_OF_MEMORY, __FUNCTION__,
        "failed to create handler");
    // LCOV_EXCL_STOP
  }

  if (__context.get() &&
      __context->GetAppState() == AppContext::APP_STATE_RUNNING) {
    __context->AddEvent(*h);
  } else {
    __pending_app_events.push_back(*h);
  }

  *event_handler = reinterpret_cast<app_event_handler_h>(h);

  return APP_ERROR_NONE;
}

extern "C" EXPORT_API int service_app_remove_event_handler(
    app_event_handler_h event_handler) {
  if (event_handler == nullptr)
    return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, nullptr);

  auto* eb = reinterpret_cast<std::shared_ptr<AppEvent>*>(event_handler);
  auto type = (*eb)->GetType();

  if (type < IAppCore::IEvent::Type::LOW_MEMORY ||
      type > IAppCore::IEvent::Type::TIME_ZONE_CHANGED ||
      type == IAppCore::IEvent::Type::DEVICE_ORIENTATION_CHANGED)
    return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__, nullptr);  // LCOV_EXCL_LINE

  if (__context.get() &&
      __context->GetAppState() == AppContext::APP_STATE_RUNNING) {
    if (!__context->RemoveEvent(*eb)) {
      // LCOV_EXCL_START
      return __on_error(APP_ERROR_INVALID_PARAMETER, __FUNCTION__,
          "invalid raw handler");
      // LCOV_EXCL_STOP
    }
  } else {
    __pending_app_events.remove(*eb);
  }

  delete eb;

  return APP_ERROR_NONE;
}

extern "C" EXPORT_API void service_app_exit_without_restart(void) {
  if (__context.get() &&
      __context->GetAppState() == AppContext::APP_STATE_RUNNING) {
    aul_status_update_v2(STATUS_NORESTART);
    __context->Exit();
  }
}

extern "C" EXPORT_API service_app_job_h service_app_add_job_handler(
    const char* job_id, service_app_job_cb callback, void* user_data) {
  if (job_id == nullptr || callback == nullptr) {
    _E("Invalid parameter");
    return nullptr;
  }

  auto* handle = new (std::nothrow) JobHandler(job_id, callback, user_data);
  if (handle == nullptr) {
    // LCOV_EXCL_START
    _E("Out of memory");
    return nullptr;
    // LCOV_EXCL_STOP
  }

  __job_manager.AddJobHandler(handle);
  __job_manager.FlushPendingJob(job_id);
  return handle;
}

extern "C" EXPORT_API int service_app_remove_job_handler(
    service_app_job_h handle) {
  if (handle == nullptr) {
    _E("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
  }

  auto* h = static_cast<JobHandler*>(handle);
  if (!__job_manager.ExistJobHandler(h)) {
    // LCOV_EXCL_START
    _E("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
    // LCOV_EXCL_STOP
  }

  __job_manager.RemoveJobHandler(h);
  delete h;
  return APP_ERROR_NONE;
}

extern "C" EXPORT_API int service_app_job_raise(int job_status,
    const char* job_id, bundle* job_data) {
  if (job_status < SERVICE_APP_JOB_STATUS_START ||
      job_status > SERVICE_APP_JOB_STATUS_STOP ||
      job_id == nullptr || job_data == nullptr) {
    _E("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
  }

  auto job = std::make_shared<Job>(job_status, job_id,
      tizen_base::Bundle(job_data, true, true));
  if (job == nullptr) {
    // LCOV_EXCL_START
    _E("Out of memory");
    return APP_ERROR_OUT_OF_MEMORY;
    // LCOV_EXCL_STOP
  }

  __job_manager.Do(job);
  return APP_ERROR_NONE;
}

extern "C" EXPORT_API int service_app_job_finished(const char* job_id) {
  if (job_id == nullptr) {
    _E("Invalid parameter");
    return APP_ERROR_INVALID_PARAMETER;
  }

  if (__job_manager.Finish(job_id) != 0)
    return APP_ERROR_INVALID_CONTEXT;  // LCOV_EXCL_LINE

  return APP_ERROR_NONE;
}
