/*
 * Copyright (c) 2020 - 2021 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gmock/gmock.h>
#include <dlog.h>
#include <glib.h>
#include <stdio.h>

#include <chrono>
#include <condition_variable>
#include <functional>
#include <mutex>
#include <thread>

#include <bundle_cpp.h>
#include <bundle_internal.h>
#include <service_app.h>
#include <service_app_extension.h>
#include <service_app_internal.h>
#include <app_common.h>
#include <aul_job_scheduler.h>
#include <glib.h>

#include "mock/test_fixture.h"
#include "mock/appcore_mock.h"

namespace appcore_agent {

class Mocks : public ::testing::NiceMock<AppCoreMock> {};

class ServiceAppTest : public TestFixture {
 public:
  ServiceAppTest() : TestFixture(std::make_unique<Mocks>()) {}
  virtual ~ServiceAppTest() {}

  virtual void SetUp() {
    context_ = g_main_context_new();
    loop_ = g_main_loop_new(context_, false);
  }

  virtual void TearDown() {
    if (loop_) {
      g_main_loop_unref(loop_);
      loop_ = nullptr;
    }

    if (context_) {
      g_main_context_unref(context_);
      context_ = nullptr;
    }
  }

  void LoopRun() {
    g_main_loop_run(loop_);
  }

  void LoopQuit() {
    g_main_loop_quit(loop_);
  }

  GMainContext* context_ = nullptr;
  GMainLoop* loop_ = nullptr;
  std::mutex mutex_;
  std::condition_variable cond_;
  bool started_ = false;
};

extern "C" int __dlog_print(log_id_t log_id, int prio, const char* tag,
    const char* fmt, ...) {
  return 0;
}

extern "C" int aul_job_scheduler_update_job_status(const char* job_id,
    aul_job_status_e job_status) {
  return 0;
}

// LCOV_EXCL_START
bool __service_app_create_cb(void* user_data) {
  return true;
}

void __service_app_terminate_cb(void* user_data) {
}

void __service_app_control_cb(app_control_h app_control, void* user_data) {
}
// LCOV_EXCL_STOP

TEST_F(ServiceAppTest, Basic) {
  // test service_app_main
  int ret = 0;
  service_app_lifecycle_callback_s callback = {
    .create = __service_app_create_cb,
    .terminate = __service_app_terminate_cb,
    .app_control = __service_app_control_cb
  };

  tizen_base::Bundle b;
  b.Add("dummy", "dummy");

  char** argv;
  int argc = bundle_export_to_argv(b.GetHandle(), &argv);
  std::unique_ptr<char**, std::function<void(char***)>> ptr(&argv,
      [argc](char*** ptr) {
    bundle_free_exported_argv(argc, ptr);
  });

  ASSERT_GT(argc, 0);
  using testing::_;

  ret = service_app_main(0, nullptr, &callback, nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  callback.create = nullptr;
  ret = service_app_main(argc, argv, &callback, nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  EXPECT_CALL(GetMock<AppCoreMock>(), Run(_, _)).Times(1);

  std::unique_lock<std::mutex> lock(mutex_);
  EXPECT_CALL(GetMock<AppCoreMock>(), OnLoopRun())
    .WillOnce(testing::Invoke([&]() {
      LoopRun();
    }));

  std::thread t([&]() {
      {
        std::unique_lock<std::mutex> t_lock(mutex_);
        GSource* source = g_timeout_source_new(100);
        g_source_set_callback(source, [](gpointer user_data) {
              auto* h = static_cast<ServiceAppTest*>(user_data);
              std::unique_lock<std::mutex> l(h->mutex_);
              h->started_ = true;
              h->cond_.notify_one();
              return G_SOURCE_REMOVE;
            }, this, nullptr);
        g_source_attach(source, context_);
        g_source_unref(source);
      }

      callback.create = __service_app_create_cb;
      service_app_main(argc, argv, &callback, this);
  });

  cond_.wait(lock, [&] { return started_; });
  ret = service_app_main(argc, argv, &callback, this);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_ALREADY_RUNNING));

  // test service_app_exit
  EXPECT_CALL(GetMock<AppCoreMock>(), Exit()).Times(1);
  EXPECT_CALL(GetMock<AppCoreMock>(), OnLoopExit())
    .WillOnce(testing::Invoke([&]() {
      LoopQuit();
      t.join();
    }));
  service_app_exit();
}

TEST_F(ServiceAppTest, Internals) {
  int ret;
  int calledCnt = 0;
  service_app_job_h handle;

  handle = service_app_add_job_handler(nullptr, nullptr, nullptr);
  EXPECT_TRUE(handle == nullptr);

  handle = service_app_add_job_handler("temp",
      [] (int status, const char* job_id, bundle* job_data,
          void* user_data) -> int {
        int* calledCnt = reinterpret_cast<int*>(user_data);
        *calledCnt = *calledCnt + 1;
        return 0;
      }, reinterpret_cast<void*>(&calledCnt));

  EXPECT_TRUE(handle != nullptr);
  EXPECT_THAT(calledCnt, testing::Eq(0));

  tizen_base::Bundle b;

  ret = service_app_job_raise(0, nullptr, nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  ret = service_app_job_raise(SERVICE_APP_JOB_STATUS_START, "temp",
      b.GetHandle());

  g_main_context_iteration(g_main_context_default(), 0);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_NONE));
  EXPECT_THAT(calledCnt, testing::Eq(1));

  ret = service_app_job_finished(nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  ret = service_app_job_finished("temp");
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_NONE));

  ret = service_app_remove_job_handler(nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  ret = service_app_remove_job_handler(handle);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_NONE));
}

TEST_F(ServiceAppTest, EventHandlers) {
  app_event_handler_h handle;
  int ret;
  int calledCnt = 0;

  ret = service_app_add_event_handler(nullptr, APP_EVENT_LOW_MEMORY, nullptr,
          nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  ret = service_app_add_event_handler(&handle, APP_EVENT_LOW_MEMORY,
            [] (app_event_info_h event_info, void* user_data) {
    int* calledCnt = reinterpret_cast<int*>(user_data);
    *calledCnt = *calledCnt + 1;
  }, reinterpret_cast<void*>(&calledCnt));

  EXPECT_THAT(ret, testing::Eq(APP_ERROR_NONE));
  EXPECT_THAT(calledCnt, testing::Eq(0));

  ret = service_app_remove_event_handler(nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  ret = service_app_remove_event_handler(handle);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_NONE));
}

TEST_F(ServiceAppTest, BasicWithoutRestart) {
  int ret = 0;
  service_app_lifecycle_callback_s callback = {
    .create = __service_app_create_cb,
    .terminate = __service_app_terminate_cb,
    .app_control = __service_app_control_cb
  };

  tizen_base::Bundle b;
  b.Add("dummy", "dummy");

  char** argv;
  int argc = bundle_export_to_argv(b.GetHandle(), &argv);
  std::unique_ptr<char**, std::function<void(char***)>> ptr(&argv,
      [argc](char*** ptr) {
    bundle_free_exported_argv(argc, ptr);
  });

  ASSERT_GT(argc, 0);
  using testing::_;

  ret = service_app_main(0, nullptr, &callback, nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  callback.create = nullptr;
  ret = service_app_main(argc, argv, &callback, nullptr);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_INVALID_PARAMETER));

  EXPECT_CALL(GetMock<AppCoreMock>(), Run(_, _)).Times(1);

  std::unique_lock<std::mutex> lock(mutex_);
  EXPECT_CALL(GetMock<AppCoreMock>(), OnLoopRun())
    .WillOnce(testing::Invoke([&]() {
      LoopRun();
    }));

  std::thread t([&]() {
      {
        std::unique_lock<std::mutex> t_lock(mutex_);
        GSource* source = g_timeout_source_new(100);
        g_source_set_callback(source, [](gpointer user_data) {
              auto* h = static_cast<ServiceAppTest*>(user_data);
              std::unique_lock<std::mutex> l(h->mutex_);
              h->started_ = true;
              h->cond_.notify_one();
              return G_SOURCE_REMOVE;
            }, this, nullptr);
        g_source_attach(source, context_);
        g_source_unref(source);
      }

      callback.create = __service_app_create_cb;
      service_app_main(argc, argv, &callback, this);
  });

  cond_.wait(lock, [&] { return started_; });
  ret = service_app_main(argc, argv, &callback, this);
  EXPECT_THAT(ret, testing::Eq(APP_ERROR_ALREADY_RUNNING));

  // test service_app_exit
  EXPECT_CALL(GetMock<AppCoreMock>(), Exit()).Times(1);
  EXPECT_CALL(GetMock<AppCoreMock>(), OnLoopExit())
    .WillOnce(testing::Invoke([&]() {
      LoopQuit();
      t.join();
    }));
  service_app_exit_without_restart();
}

}  // namespace appcore_agent

int main(int argc, char* argv[]) {
  int ret = 0;
  try {
    ::testing::InitGoogleTest(&argc, argv);
  } catch(...) {
    // LCOV_EXCL_START
    std::cout << "Exception occured" << std::endl;
    return 1;
    // LCOV_EXCL_STOP
  }

  try {
    return RUN_ALL_TESTS();
  } catch(const ::testing::internal::GoogleTestFailureException& e) {
    // LCOV_EXCL_START
    std::cout << "GoogleTestFailureException occured:" << e.what() << std::endl;
    ret = 1;
    // LCOV_EXCL_STOP
  }

  return ret;
}
